""" Launch commands

These commands meant as development tools only
"""
import os
import click

from flask import current_app as app
from halo.halo import Halo

from covfee.server.start import create_app
from covfee.cli.utils import working_directory
from covfee.shared.schemata import Schemata

@click.group(name='covfee-dev')
def covfee_dev_cli():
    pass


@covfee_dev_cli.command(name='build')
def build_master():
    """
    Builds the master (without custom tasks) for distribution.
    """
    _, app = create_app('dev')

    with app.app_context():
        bundle_path = app.config['MASTER_BUNDLE_PATH']
        with working_directory(app.config['COVFEE_CLIENT_PATH']):
            os.system('npx webpack' +
                    ' --config ./webpack.prod.js' + ' --output-path '+bundle_path)

@covfee_dev_cli.command(name='schemata')
def make_schemata():
    _, app = create_app('dev')

    with app.app_context():
        with Halo(text='Making schemata', spinner='dots') as spinner:
            schema = Schemata()
            schema.make()
            spinner.succeed('Schemata made.')



import os
import json
import subprocess

import zmq

from covfee.cli.utils import working_directory
from flask import current_app as app
context = zmq.Context()

class ReduxStoreService:
    def run(self):
        with working_directory(os.path.join(app.config['COVFEE_SERVER_PATH'], 'socketio')):
            subprocess.Popen(['npx', 'pm2', 'start', 'reduxStore.js', '-i', '1', '--watch', '--', 
                'serve',
                '--database',
                app.config['DATABASE_PATH']
            ])

class ReduxStoreClient:
    ''' This class takes care of replicating the shared state in multi-party tasks server-side for 
    persistence and synchronization.
    Multiparty tasks use a synced redux store for state. The server maintains the true state by dispatching
    the Redux actions sent by each client. After being executed on the server's store, actions are sent back
    to all clients to update their state to match the server's true state.
    The true state is stored in the database for persistence.
    The server's state is kept in a nodejs service that this class communicates with via zmq.
    '''

    def __init__(self):
        #  Socket to talk to server
        self.socket = context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.RCVTIMEO, 500)
        self.socket.connect("tcp://127.0.0.1:5555")

    def socket_request(self, payload):
        self.socket.send_json(payload)
        message = self.socket.recv()
        return json.loads(message.decode('utf-8'))

    def join(self, responseId):
        return self.socket_request({
            'command': 'join', 
            'responseId': responseId
        })

    def leave(self, responseId):
        return self.socket_request({
            'command': 'leave', 
            'responseId': responseId
        })

    def action(self, responseId, action):
        return self.socket_request({
            'command': 'action', 
            'responseId': responseId,
            'action': action
        })

    def state(self, responseId):
        return self.socket_request({
            'command': 'state', 
            'responseId': responseId
        })
    
    def reset(self, responseId):
        return self.socket_request({
            'command': 'reset', 
            'responseId': responseId
        })



import os
from pathlib import Path
import contextlib
from shutil import which


class NPMPackage:
    def __init__(self, path):
        self.path = path

    def is_npm_available(self):
        return which('npm') is not None

    def is_installed(self):
        return os.path.exists(os.path.join(self.path, 'node_modules'))

    def install(self):
        with working_directory(self.path):
            os.system('npm install')


@contextlib.contextmanager
def working_directory(path):
    """Changes working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)



class VideoRecallTask:
    def __init__(self):
        pass

    @staticmethod
    def aggregate_chunks(chunks):
        # concatenate all chunks together
        return [x for y in chunks for x in y]



from .base import BaseCovfeeTask


class Continuous1DTask(BaseCovfeeTask):

    def aggregate_chunks(self, chunks):
        # concatenate all chunks together
        return [x for y in chunks for x in y]



import os
from urllib.parse import urlparse

from .db import db
from .hit import *
from .project import *
from .task import *
from .user import *


def load_config(app, mode):
    app.config['COVFEE_ENV'] = mode
    # load the base configuration object
    app.config.from_object('covfee.server.config')

    # update with the custom project config files
    if mode == 'local':
        app.config.from_pyfile(os.path.join(
            os.getcwd(), 'covfee.local.config.py'), silent=True)
    elif mode == 'dev':
        app.config.from_pyfile(os.path.join(
            os.getcwd(), 'covfee.development.config.py'), silent=True)
    elif mode == 'deploy':
        app.config.from_pyfile(os.path.join(
            os.getcwd(), 'covfee.deployment.config.py'), silent=True)
    else:
        raise Exception(f'Unrecognized application mode {mode}.')

    # check if SSL enabled
    app.config['SSL_ENABLED'] = ('SSL_KEY_FILE' in app.config and 'SSL_CERT_FILE' in app.config)

    # apply extended config
    app_path = urlparse(app.config['BASE_URL']).path
    if app_path == '':
        app_path = '/'

    app.config.update(
        # copy over secret key
        JWT_SECRET_KEY=app.config['COVFEE_SECRET_KEY'],

        # create sqlalchemy database uri
        SQLALCHEMY_DATABASE_URI=f'sqlite:///{app.config["DATABASE_PATH"]}',

        # create derived URLs
        PROJECT_WWW_URL=app.config.get('PROJECT_WWW_URL', app.config['BASE_URL'] + '/www'),
        BUNDLE_URL=app.config.get('BUNDLE_URL', app.config['BASE_URL'] + '/www'),
        APP_URL=app.config['BASE_URL'] + '/#',
        ADMIN_URL=app.config['BASE_URL'] + '/admin#',
        LOGIN_URL=app.config['BASE_URL'] + '#login',
        API_URL=app.config['BASE_URL'] + '/api',
        AUTH_URL=app.config['BASE_URL'] + '/auth',
        # Set the cookie paths, so that you are only sending your access token
        # cookie to the access endpoints, and only sending your refresh token
        # to the refresh endpoint. Technically this is optional, but it is in
        # your best interest to not send additional cookies in the request if
        # they aren't needed.
        JWT_ACCESS_COOKIE_PATH=app_path,
        JWT_REFRESH_COOKIE_PATH=os.path.join(app_path, 'auth/refresh')
    )

    # point to webpack-dev-server bundles in dev mode
    if mode == 'dev':
        app.config['BUNDLE_URL'] = app.config['DEV_BUNDLE_URL']

    # create the frontend config object:
    app.config['FRONTEND_CONFIG'] = {
        # frontend only has two environments: production and development
        'env': 'production' if mode in ['local', 'deploy'] else 'development',
        'socketio_enabled': app.config['SOCKETIO_ENABLED'],
        'google_client_id': app.config['GOOGLE_CLIENT_ID'],
        'www_url': app.config['PROJECT_WWW_URL'],
        'app_url': app.config['APP_URL'],
        'admin_url': app.config['ADMIN_URL'],
        'base_url': app.config['BASE_URL'],
        'api_url': app.config['API_URL'],
        'auth_url': app.config['AUTH_URL'],
    }



import os

# Configuration file specifying the default covfee behavior
# Can be overwritten via project config

# GENERAL SETTINGS
BASE_URL = 'http://localhost:5000'


#  Security: Used as salt used for generating hash links and for json web tokens (JWT_SECRET_KEY)
COVFEE_SECRET_KEY = 'CHANGE_ME'

SOCKETIO_ENABLED = True
RTSTORE_ENABLED = True

GOOGLE_CLIENT_ID = None

# sqlalchemy settings
SQLALCHEMY_ENGINE_OPTIONS = {'isolation_level': "READ UNCOMMITTED"}
SQLALCHEMY_TRACK_MODIFICATIONS = False

COVFEE_BASE_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')
COVFEE_CLI_PATH = os.path.join(COVFEE_BASE_PATH, 'cli')
COVFEE_CLIENT_PATH = os.path.join(COVFEE_BASE_PATH, 'client')
COVFEE_SERVER_PATH = os.path.join(COVFEE_BASE_PATH, 'server')
SHARED_PATH = os.path.join(COVFEE_BASE_PATH, 'shared')
# for storing the json schemata for validation
SCHEMATA_PATH = os.path.join(COVFEE_BASE_PATH, 'shared', 'schemata.json')

# Configure application to store JWTs in cookies. Whenever you make
# a request to a protected endpoint, you will need to send in the
# access or refresh JWT via a cookie.
JWT_TOKEN_LOCATION = ['cookies']
JWT_COOKIE_CSRF_PROTECT = False

# PROJECT FOLDER PATHS

# database
DATABASE_RELPATH = os.path.join(os.getcwd(), '.covfee', 'database.covfee.db')
DATABASE_PATH = os.path.join(os.getcwd(), DATABASE_RELPATH)

# project www and bundle location
PROJECT_WWW_PATH = os.path.join(os.getcwd(), 'www')

# enables the www server
SERVE_WWW = True

MASTER_BUNDLE_PATH = os.path.join(COVFEE_BASE_PATH, 'client', 'bundles')

# for temporary (download) files
TMP_PATH = os.path.join(os.getcwd(), 'tmp')

# DEV MODE SETTINGS
DEV_BUNDLE_URL = 'http://localhost:8085'





import datetime
from hashlib import sha256
import os
import json

import pandas as pd
from flask import current_app as app

from .db import db
from .hit import HIT


class Project(db.Model):
    """ Represents a set of HITs which make up an experiment or annotation project """
    __tablename__ = 'projects'

    id = db.Column(db.LargeBinary, primary_key=True)
    name = db.Column(db.String)
    email = db.Column(db.String)
    hits = db.relationship("HIT", backref="project", cascade="all, delete-orphan")

    created_at = db.Column(db.DateTime, default=datetime.datetime.now)
    updated_at = db.Column(db.DateTime, onupdate=datetime.datetime.now)

    def __init__(self, id, name, email, hits, **kwargs):
        hashstr = Project.get_hashtr(id)
        self.id = sha256(hashstr.encode()).digest()
        self.name = name
        self.email = email

        hashstr = Project.get_hashtr(id)
        for hit_dict in hits:
            hit = HIT(**hit_dict, project_id=id)
            self.hits.append(hit)

            repeats = hit_dict.get('repeat', 1)
            for r in range(repeats):
                hit.instantiate()

    def get_dataframe(self):
        list_of_instances = list()
        for hit in self.hits:
            for instance in hit.instances:
                list_of_instances.append({
                    'hit_name': hit.name,
                    'id': instance.id.hex(),
                    'url': instance.get_url(),
                    'preview_url': instance.get_preview_url(),
                    'completion_code': instance.get_completion_code()
                })
        df = pd.DataFrame(list_of_instances, columns=['hit_name', 'id', 'url', 'preview_url', 'completion_code'])
        
        return df

    def as_dict(self, with_hits=False, with_instances=False, with_config=False):
        project_dict = {c.name: getattr(self, c.name)
                        for c in self.__table__.columns}
        project_dict['id'] = project_dict['id'].hex()
        if with_hits:
            project_dict['hits'] = [hit.as_dict(with_instances=with_instances, with_config=with_config) for hit in self.hits]
        return project_dict

    def info(self):
        txt = f'{self.name}\nID: {self.id.hex():s}\n'
        for tl in self.hits:
            txt += tl.showinfo()
        return txt

    def __str__(self):
        txt = f'{self.name} - ID: {self.id.hex():s}\n'
        for tl in self.hits:
            txt += str(tl)
        return txt

    @staticmethod
    def get_hashtr(id):
        return id + app.config['COVFEE_SECRET_KEY']

    @staticmethod
    def get_id(id):
        return sha256(Project.get_hashtr(id).encode()).digest()

    @staticmethod
    def from_json(fpath: str):
        '''
        Loads a project into ORM objects from a project json file.
        '''
        with open(fpath, 'r') as f:
            proj_dict = json.load(f)

        return Project(**proj_dict)

    def stream_download(self, z, base_path, submitted_only=True, csv=False):
        for hit in self.hits:
            for instance in hit.instances:
                if submitted_only and not instance.submitted:
                    continue
                yield from instance.stream_download(z, os.path.join(base_path, instance.id.hex()), csv=csv)



import os
import time
import signal
import subprocess

from flask import current_app as app
from ..orm import db

from .base import BaseCovfeeTask
from ..utils.agora.rtc_token_builder import RtcTokenBuilder, Role_Attendee

class VideocallTask(BaseCovfeeTask):
    def get_task_specific_props(self) -> dict:

        if 'AGORA_APPID' not in app.config or 'AGORA_CERT' not in app.config:
            return {}

        # token valid for 10 hours
        privilegeExpiredTs = int(time.time()) + 10*60*60 # in seconds
        return {
            'agoraAppId': app.config['AGORA_APPID'],
            'agoraToken': RtcTokenBuilder.buildTokenWithUid(app.config['AGORA_APPID'], app.config['AGORA_CERT'], str(self.task.id), 0, Role_Attendee, privilegeExpiredTs)
        }

    def on_first_join(self):
        ''' Starts the recording script
        '''
        recording_dir = os.path.join(app.config["PROJECT_WWW_PATH"], self.task.hitinstance.id.hex(), f'{self.task.id}_{str(self.response.id)}')
        if not os.path.exists(recording_dir):
            os.makedirs(recording_dir)

        cmd =  [f'{app.config["AGORA_RECORDER_PATH"]}',
                f' --appId {app.config["AGORA_APPID"]}',
                f' --channel {self.response.id}',
                f' --recordFileRootDir {recording_dir}',
                f' --triggerMode 1',
                f' --uid 0',
                f' --channelProfile 0',
                f' --appliteDir {app.config["AGORA_APPLITEDIR"]}']

        p = subprocess.Popen(cmd)
        print(' '.join(cmd))
        self.response.extra['recorder_pid'] = p.pid
        db.session.commit()
        # os.system(f'kill -s 10 {p.pid}')
        os.kill(p.pid, 10)


    def on_last_leave(self):
        ''' Stops the recording script
        '''
        if 'recorder_pid' in self.response.extra:
            os.kill(self.response.extra['recorder_pid'], 12)
            del self.response.extra['recorder_pid']
            db.session.commit()

